/* FourWays - a BeOS imput_server add-in enabling gesture recognition
 *
 * FourWaysFilter.cpp
 * 
 * Copyright (C) 2001 Stefan Werner (stefan@keindesign.de)
 *
 * 	FindWindow function used with permission from Thomas Thiriez
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <string.h>
#include <malloc.h>
#include <stdlib.h>

#include "Messenger.h"

#include <Application.h>
#include <Roster.h>
#include <Message.h>
#include <InterfaceDefs.h>
#include <Window.h>
#include <Alert.h>

#include "FourWaysFilter.h"
#include "interface.h"
extern "C" {
	#include "stroke.h"
}

#define HOTBUTTON B_SECONDARY_MOUSE_BUTTON
//#define HOTBUTTON B_TERTIARY_MOUSE_BUTTON

const bigtime_t HUGE_TIMEOUT = 1000000;

FourWaysFilter::FourWaysFilter() :
	mWinMsgr(0),
	mValidFrame(false),
	mValidLook(false),
	mIsMoving(false),
	init(B_OK)
{
	validClick = false;
	return;
}

FourWaysFilter::~FourWaysFilter()
{
	if(mWinMsgr)
		delete mWinMsgr;
}

status_t FourWaysFilter::InitCheck()
{
	return B_OK;
}

uint32 FourWaysFilter::KeyFromSequence (char *sequence)
{
	uint32 result = strtoul (sequence, NULL, 16) & 0x7fffffff;
	return result;
}

filter_result FourWaysFilter::Filter(BMessage *message, BList *outList)
{
	BPoint where;
	filter_result ret=B_DISPATCH_MESSAGE;
	
	switch(message->what)
	{
		case B_MOUSE_DOWN:
			if (message->FindInt32 ("buttons", &buttons) != B_OK) break;
			if (message->FindPoint ("where" , &where) != B_OK) break;
			if (buttons & HOTBUTTON)
			{
				tempMessage = new BMessage (*message);	// remember the message, maybe we need to resend it later	
				stroke_init ();							// record the stroke and add the first point
				stroke_record ((int)where.x, (int)where.y);
				ret = B_SKIP_MESSAGE;					// don't pass this message
				validClick = true;						// track further mouse movements
				if(!mWinMsgr)
					mWinMsgr=FindWindow(where);			// get a BMessenger for the window the pointer points at
			}
			break;
			
		case B_MOUSE_MOVED:
			if (validClick)								// parse only if this is a valid stroke
				if (message->FindPoint ("where" , &where) != B_OK) 
					break;								// should never occur, but we want to be on the safe side
				else									// so let's add this point
					stroke_record ((int)where.x, (int)where.y);
			break;
			
		case B_MOUSE_UP:
			if (validClick)								// parse only if this is a valid stroke
			{
				#define THRESHOLD 70
				
				if (message->FindPoint ("where" , &where) != B_OK) break;
				int delta_x = int (where.x - mMousePos.x);
				int delta_y = int (where.y - mMousePos.y);
				
				if ((abs(delta_x) > THRESHOLD) || (abs(delta_y) > THRESHOLD))
				{
					char sequence[255];
					stroke_trans (sequence);			// get the result from libstroke
					
					if (strcmp (sequence, "14789") == 0)
					{									// down-right will close the window
						if (mWinMsgr)
							mWinMsgr->SendMessage (B_QUIT_REQUESTED);
						ret = B_SKIP_MESSAGE;
					}
					
					if (strcmp (sequence, "852") == 0)
					{									// up to zoom
						if (mWinMsgr)
							mWinMsgr->SendMessage (B_ZOOM);
						ret = B_SKIP_MESSAGE;
					}
				
					if (strcmp (sequence, "456") == 0)
					{									// left and right to cahnge the current workspace
						if (current_workspace() < (count_workspaces() - 1))
							activate_workspace (current_workspace() + 1);
						else
							activate_workspace (0);
						ret = B_SKIP_MESSAGE;
					}
					
					if (strcmp (sequence, "654") == 0)
					{
						if (current_workspace() > 0)
							activate_workspace (current_workspace() - 1);
						else
							activate_workspace (count_workspaces () - 1);						
						ret = B_SKIP_MESSAGE;
					}
					if (strcmp (sequence, "258") == 0)
					{
						if (mWinMsgr)
						{
							BMessage *new_message = new BMessage (B_MINIMIZE);
							new_message->AddBool ("minimize", false);
							mWinMsgr->SendMessage (new_message);
							delete new_message;
						}
						ret = B_SKIP_MESSAGE;
					}
					if (ret != B_SKIP_MESSAGE)
					{
						message->what = B_UNMAPPED_KEY_DOWN;
						message->AddInt32("key", 0x80000000 | KeyFromSequence (sequence));
						message->AddString("stroke:phonepad", sequence);
						message->ReplaceInt64("when", system_time());
						ret = B_DISPATCH_MESSAGE;
					}
				} else { 								// not a recognized stroke, so pass the mouse events on
					
					if (ret != B_SKIP_MESSAGE)
					{
						BMessage *new_message = new BMessage (*message);
					//if (tempMessage)
						tempMessage->ReplaceInt64("when", system_time());
						outList->AddItem (tempMessage);
						outList->AddItem (new_message);
						ret = B_DISPATCH_MESSAGE;
					}
				}
				validClick = false;

				if(mWinMsgr)
				{
					mWinMsgr = 0;
				}
			}
			break;
	};
	
	return ret;
}

/* 	FindWindow function
	returns a Messenger for the Window under the mouse pointer
	used with permission from Thomas Thiriez
*/

BMessenger *FourWaysFilter::FindWindow(BPoint pos)
{
	int32 *tokens;
	int32 count;
	
	tokens=get_token_list(-1, &count);
	if(!tokens)
		return 0;
	
	for(int32 i=0; i<count; i++)
	{
		window_info *wi;
		wi=get_window_info(tokens[i]);
		if(!wi)
			continue;
		if(wi->layer<=2)
		{
			free(wi);
			free(tokens);
			return 0;
		}
		if((wi->window_left<=pos.x) && (wi->window_right>=pos.x) &&
		   (wi->window_top<=pos.y) && (wi->window_bottom>=pos.y))
		{
			BMessenger *msgr=new BMessenger;
			msgr->fPort=wi->client_port;
			msgr->fHandlerToken=wi->client_token;
			msgr->fTeam=wi->team;
			mWinToken=tokens[i];
			mWinWorkspaceMask=wi->workspaces;
			mWinFlags=wi->flags;
			mIsFrontMost=!i;
			
			if(i<count-1)
			{
				window_info *wi2=get_window_info(tokens[i+1]);
				mIsBackMost=wi2->layer<=2;
				free(wi2);
			}
			else
				mIsBackMost=false;
			
			free(wi);
			free(tokens);
			
			BMessage msg(B_GET_PROPERTY);
			BMessage reply;
			msg.AddSpecifier("Frame");
			msgr->SendMessage(&msg, &reply, HUGE_TIMEOUT, HUGE_TIMEOUT);
			mValidFrame=(reply.FindRect("result", &mWinFrame)==B_OK);
			
			BMessage msg2(B_GET_PROPERTY);
			BMessage reply2;
			msg2.AddSpecifier("Look");
			msgr->SendMessage(&msg2, &reply2, HUGE_TIMEOUT, HUGE_TIMEOUT);
			mValidLook=(reply2.FindInt32("result", (int32 *)&mWinLook)==B_OK);
			return msgr;
		}
		free(wi);
	}
	
	free(tokens);
	return 0;
}
